from dataclasses import dataclass

import typer


@dataclass
class IdLangPair:
    track_id: int
    lang: str

    @classmethod
    def parse(cls, value: str) -> "IdLangPair":
        try:
            track_id_str, lang = value.split(":", 1)
        except ValueError as e:
            raise typer.BadParameter(
                f"Invalid format: '{value}'. Expected '<track_id>:<lang>'."
            ) from e
        return cls(int(track_id_str), lang)



from .add_subs import AddSubtitlesAction
from .base import Action, ActionContext
from .set_title import SetTitleAction
from .set_track_lang import SetTrackLanguageAction

__all__ = [
    "Action",
    "ActionContext",
    "AddSubtitlesAction",
    "SetTitleAction",
    "SetTrackLanguageAction",
]



"""Global operations (affect entire file)."""

from ..mkv import MKVFile
from . import Operation


class SetTitle(Operation):
    """Set the title of an MKV file."""

    def __init__(self, title: str):
        self.title = title

    def apply(self, video: MKVFile):
        video.title = self.title



"""Set track language operation."""

import typer

from ..mkv import MKVFile
from . import Operation


class SetTrackLanguage(Operation):
    """Set the language of a specific track in an MKV file."""

    def __init__(self, track_id: int, lang: str):
        self.track_id = track_id
        self.lang = lang

    def apply(self, video: MKVFile):
        if track := video.select_track(self.track_id):
            track.language = self.lang
            track.language_ietf = self.lang
        else:
            raise typer.BadParameter(
                f"Track {self.track_id} not found in file {video.file_path}"
            )



from pathlib import Path

import typer
from rich.console import Console
from rich.table import Table

from ..mkv import MkvInfo, MKVWrapper
from ..utils.files import collect_video_files
from .common import QuietOption


def info(
    target: Path = typer.Argument(
        Path("."),
        help="Path to the video file or directory containing video files.",
        exists=True,
        readable=True,
        resolve_path=True,
    ),
    quiet: bool = QuietOption,
):
    """Display information about the specified video file or all video files in the given directory."""
    console = Console(quiet=quiet)
    files = collect_video_files(target)

    if not files:
        console.print(f"[yellow]No video files found in {target}[/yellow]")
        raise typer.Exit(0)

    wrapper = MKVWrapper(console)
    videos = wrapper.load_videos(files)

    _print_table([video.info for video in videos])


def _print_table(videos: list[MkvInfo]):
    console = Console()
    table = Table("File", "Title", "Tracks")
    for video in videos:
        tt = Table("ID", "Type", "Codec", "Lang", "Name", "Default", "Forced")
        for track in video.tracks:
            tt.add_row(
                str(track.id),
                track.type or "-",
                track.codec or "-",
                track.lang or "-",
                track.name or "-",
                "Yes" if track.default else "No",
                "Yes" if track.forced else "No",
            )
        table.add_row(video.path.name, video.title or "-", tt)
    console.print(table)



from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

import typer
from rich.console import Console

from .actions import Action, ActionContext
from .mkv import MKVFile, MKVWrapper
from .operations import Operation


@dataclass
class FilePlan:
    video: MKVFile
    output_path: Path


class ProcessingEngine:
    def __init__(self, console: Console):
        self.console = console
        self.mkv_wrapper = MKVWrapper(console)

    def _collect_operations(
        self, target: Path, actions: list[Action]
    ) -> dict[Path, list[Operation]]:
        context = ActionContext(target, self.console)
        reduced = defaultdict[Path, list[Operation]](list)

        for action in actions:
            for file_path, operations in action.analyze(context).items():
                reduced[file_path].extend(operations)
        return reduced

    def plan(self, target: Path, actions: list[Action]) -> list[FilePlan]:
        """Process the target with the given actions."""
        operations_map = self._collect_operations(target, actions)
        mkv_files = self.mkv_wrapper.load_videos(list(operations_map.keys()))

        for mkv_file in mkv_files:
            for operation in operations_map[mkv_file.file_path]:
                operation.apply(mkv_file)
        return [
            FilePlan(
                video=video, output_path=self._compute_output_path(video.file_path)
            )
            for video in mkv_files
            if video.changed
        ]

    def _compute_output_path(self, source: Path) -> Path:
        return source.with_suffix(source.suffix + ".tmp")

    def execute(self, plans: list[FilePlan]):
        """Execute the given plans."""
        codes = self.mkv_wrapper.mux_videos(
            [(plan.output_path, plan.video) for plan in plans]
        )
        success_count = sum(1 for code in codes if code == 0)
        error_count = len(codes) - success_count
        color = (
            "green" if error_count == 0 else "yellow" if success_count > 0 else "red"
        )
        self.console.print(
            f"[{color}]Processing complete:[/] {success_count} succeeded, {error_count} failed."
        )
        if not success_count:
            raise typer.Exit(code=1)



import os
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache
from pathlib import Path

from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn

from .models import MKVFile

os.environ["LANG"] = "en_US.UTF-8"  # Ensure consistent language for mkvtoolnix output
console = Console()


class MKVWrapper:
    def __init__(self, console: Console):
        self.console = console

    def load_videos(self, file_paths: list[Path]) -> list[MKVFile]:
        """Load multiple MKV files concurrently with caching and progress indication."""
        with (
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                console=self.console,
            ) as progress,
            ThreadPoolExecutor() as executor,
        ):

            def load_video(file_path: Path) -> MKVFile:
                progress.add_task(f"Loading {file_path.name}...", total=None)
                return _load_mkv(file_path)

            return list(executor.map(load_video, file_paths))

    def mux_videos(self, videos: list[tuple[Path, MKVFile]]) -> list[int]:
        """Mux multiple MKV files concurrently with progress indication."""
        with (
            Progress(console=self.console) as progress,
            ThreadPoolExecutor() as executor,
        ):

            def mux_video(arg: tuple[Path, MKVFile]) -> int:
                path, video = arg
                id = progress.add_task(f"Muxing {path.name}...", total=100)

                def progress_handler(progress_value: int):
                    progress.update(id, completed=progress_value)

                code = video.mux(  # pyright: ignore[reportUnknownMemberType]
                    output_path=path,
                    progress_handler=progress_handler,
                )
                video.file_path.unlink()
                path.rename(video.file_path)
                return code

            return list(executor.map(mux_video, videos))


@lru_cache(maxsize=32)
def _load_mkv(file_path: Path) -> MKVFile:
    return MKVFile(file_path)



import typer

from muxwell import __version__

from .apply import apply
from .info import info

app = typer.Typer(
    name="muxwell",
    help="A command-line tool for managing MKV files and subtitles with batch operations support.",
    no_args_is_help=True,
)


def version_callback(value: bool):
    if value:
        typer.echo(__version__)
        raise typer.Exit()


@app.callback()
def main(
    version: bool = typer.Option(
        False, "--version", callback=version_callback, help="Show version and exit."
    ),
):
    """Muxwell: A command-line tool for managing MKV files and subtitles with batch operations support."""
    pass


app.command(help="Show info about video file(s).")(info)
app.command(help="Apply changes to video file(s).", no_args_is_help=True)(apply)



"""Set the language of a specific track in the MKV file."""

from ..operations import SetTrackLanguage
from ..utils.lang import normalize_language
from .base import Action, ActionContext


class SetTrackLanguageAction(Action):
    """Set the language of a specific track in the MKV file."""

    def __init__(self, track_id: int, lang: str):
        self.track_id = track_id
        self.lang = normalize_language(lang)

    def analyze(self, context: ActionContext):
        """Create SetTrackLanguage operation with the provided language."""
        return {
            v: [SetTrackLanguage(self.track_id, self.lang)] for v in context.video_files
        }



"""Base Operation class."""

from abc import ABC, abstractmethod

from ..mkv import MKVFile


class Operation(ABC):
    """Base class for all MKV operations."""

    @abstractmethod
    def apply(self, video: MKVFile):
        """Apply the operation to an MKV file."""
        pass



from pathlib import Path

import typer
from rich.console import Console

from ..actions import Action, AddSubtitlesAction, SetTitleAction, SetTrackLanguageAction
from ..engine import ProcessingEngine
from .common import QuietOption
from .models import IdLangPair


def apply(
    target: Path = typer.Argument(
        Path("."),
        help="Path to the video file or directory containing video files.",
        exists=True,
        readable=True,
        resolve_path=True,
    ),
    set_title: str | None = typer.Option(
        None, "--set-title", help="Set the title of the MKV file."
    ),
    add_subs: list[Path] = typer.Option(
        [],
        help="Path to subtitles file. Can be used multiple times.",
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    ),
    set_track_lang: list[IdLangPair] = typer.Option(
        [],
        parser=IdLangPair.parse,
        help="Set the language of a specific track in the format <track_id>:<lang>. Can be used multiple times.",
    ),
    quiet: bool = QuietOption,
):
    """Process the specified video file or all video files in the given directory."""
    console = Console(quiet=quiet)

    actions: list[Action] = []
    if set_title:
        actions.append(SetTitleAction(set_title))

    for sub_path in add_subs:
        actions.append(AddSubtitlesAction(sub_path))

    for track_lang in set_track_lang:
        actions.append(SetTrackLanguageAction(track_lang.track_id, track_lang.lang))

    engine = ProcessingEngine(console)
    plans = engine.plan(target, actions)
    if not plans:
        console.print("[yellow]No changes to apply.[/yellow]")
        raise typer.Exit(0)
    engine.execute(plans)



"""Add subtitles track operation."""

from pathlib import Path

from pymkv import MKVTrack

from ..mkv import MKVFile
from .base import Operation


class AddSubtitles(Operation):
    """Add a subtitles track to an MKV file."""

    def __init__(self, path: Path):
        self.path = path

    def apply(self, video: MKVFile):
        video.add_track(MKVTrack(str(self.path)))



from .add_subs import AddSubtitles
from .base import Operation
from .global_ops import SetTitle
from .set_track_lang import SetTrackLanguage

__all__ = ["AddSubtitles", "Operation", "SetTitle", "SetTrackLanguage"]



from pathlib import Path

VIDEO_EXTENSIONS = [".mkv", ".mp4", ".m4v", ".avi", ".mov", ".wmv", ".flv"]
SUBTITLES_EXTENSIONS = [".srt", ".ass", ".ssa"]


def collect_files(target: Path, extensions: list[str]) -> list[Path]:
    """Collect all files with the given extensions in the specified directory."""
    if target.is_file():
        if any(target.suffix.lower() == ext for ext in extensions):
            return [target]
        else:
            raise ValueError(
                f"The file {target} does not have a supported video extension."
            )

    files: list[Path] = []
    for ext in extensions:
        files.extend(target.glob(f"*{ext}"))
    return files


def collect_video_files(target: Path) -> list[Path]:
    """Collect all video files in the specified directory."""
    return collect_files(target, VIDEO_EXTENSIONS)



"""Base Action classes."""

from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

import typer
from rich.console import Console

from ..mkv import MKVWrapper
from ..operations import Operation
from ..utils.files import collect_video_files


@dataclass
class ActionContext:
    """Context available during actions analysis."""

    target: Path
    console: Console

    @cached_property
    def mkv_wrapper(self) -> MKVWrapper:
        return MKVWrapper(self.console)

    @cached_property
    def video_files(self) -> list[Path]:
        if videos := collect_video_files(self.target):
            return videos
        self.console.print(f"[yellow]No video files found in {self.target}[/yellow]")
        raise typer.Exit(0)


class Action(ABC):
    """Base class for high-level actions."""

    @abstractmethod
    def analyze(self, context: ActionContext) -> Mapping[Path, Sequence[Operation]]:
        """
        Analyze the context and return operations for each file.

        Returns:
            Dict mapping file path -> list of operations to apply
        """
        pass



"""Action to add a subtitles track to an MKV file."""

from pathlib import Path

import typer

from ..operations import AddSubtitles
from ..utils.files import SUBTITLES_EXTENSIONS
from .base import Action, ActionContext


class AddSubtitlesAction(Action):
    """Add subtitles track"""

    def __init__(self, path: Path):
        if path.suffix.lower() not in SUBTITLES_EXTENSIONS:
            raise typer.BadParameter(
                f"Unsupported subtitles format: {path.suffix}, supported formats are: {', '.join(SUBTITLES_EXTENSIONS)}"
            )
        self.path = path

    def analyze(self, context: ActionContext):
        """Create AddSubtitles operation with the provided path."""
        return {v: [AddSubtitles(self.path)] for v in context.video_files}



import typer
from iso639 import Language, LanguageNotFoundError


def normalize_language(code: str) -> str:
    """
    Normalize a language identifier to a 3-letter ISO-639-2/B code.

    :param code: A language identifier (e.g., 'en', 'eng', 'fra', 'fre', 'French' etc.)
    :type code: str
    :return: A normalized 3-letter ISO-639-2/B language code.
    :rtype: str
    :raises typer.BadParameter: if the language cannot be resolved.
    """
    try:
        lang = Language.match(code, strict_case=False)
        if not lang.part2b:
            raise typer.BadParameter(f"Invalid language code: {code}")
        return lang.part2b
    except LanguageNotFoundError as e:
        raise typer.BadParameter(f"Invalid language code: {code}") from e



"""Infer title from filename action."""

from ..operations import SetTitle
from . import Action, ActionContext


class SetTitleAction(Action):
    """Set title"""

    def __init__(self, title: str):
        self.title = title

    def analyze(self, context: ActionContext):
        """Create SetTitle operation with the provided title."""
        return {v: [SetTitle(self.title)] for v in context.video_files}



from dataclasses import dataclass
from pathlib import Path
from typing import cast

from pymkv import MKVFile as PyMKVFile, MKVTrack as PyMKVTrack


@dataclass
class TrackInfo:
    id: int
    type: str | None
    codec: str | None
    lang: str | None
    name: str | None
    default: bool
    forced: bool


@dataclass
class MkvInfo:
    path: Path
    title: str | None
    tracks: list[TrackInfo]


class MKVFile(PyMKVFile):
    """Extended MKVFile class for additional functionalities."""

    def __init__(self, file_path: str | Path):
        super().__init__(file_path)  # pyright: ignore[reportUnknownMemberType]
        self.file_path = Path(file_path)
        self._snapshot = self.info

    @property
    def info(self) -> MkvInfo:
        """Get structured information about the MKV file."""
        return MkvInfo(
            path=self.file_path,
            title=self.title,
            tracks=[
                TrackInfo(
                    id=track.track_id,
                    type=track.track_type,
                    codec=track.track_codec,
                    lang=track.language,
                    name=track.track_name,
                    default=track.default_track or False,
                    forced=track.forced_track or False,
                )
                for track in self.tracks
            ],
        )

    @property
    def changed(self) -> bool:
        """Check if the MKV file was modified."""
        return self.info != self._snapshot

    def select_track(self, track_id: int) -> PyMKVTrack | None:
        """Get a track by its ID."""
        try:
            return cast(PyMKVTrack, self.get_track(track_id))
        except IndexError:
            return None
